# Generated from grammar/DynamoDbGrammar.g4 by ANTLR 4.7.2
# encoding: utf-8
# pylint: disable=W
from antlr4 import *
from io import StringIO
from typing.io import TextIO # pylint: disable=import-error
import sys


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3 ")
        buf.write("\u00f1\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
        buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
        buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
        buf.write("\t\31\4\32\t\32\4\33\t\33\3\2\3\2\3\2\3\3\3\3\3\3\7\3")
        buf.write("=\n\3\f\3\16\3@\13\3\3\4\3\4\3\4\3\5\3\5\3\5\3\5\3\5\3")
        buf.write("\5\3\5\3\5\3\5\3\5\3\5\7\5P\n\5\f\5\16\5S\13\5\3\5\3\5")
        buf.write("\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3")
        buf.write("\5\5\5e\n\5\3\5\3\5\3\5\3\5\3\5\3\5\7\5m\n\5\f\5\16\5")
        buf.write("p\13\5\3\6\3\6\3\7\3\7\3\7\3\b\3\b\3\b\3\b\6\b{\n\b\r")
        buf.write("\b\16\b|\3\t\3\t\3\t\3\t\7\t\u0083\n\t\f\t\16\t\u0086")
        buf.write("\13\t\3\n\3\n\3\n\3\n\3\13\3\13\3\13\3\13\7\13\u0090\n")
        buf.write("\13\f\13\16\13\u0093\13\13\3\f\3\f\3\f\3\r\3\r\3\r\3\r")
        buf.write("\7\r\u009c\n\r\f\r\16\r\u009f\13\r\3\16\3\16\3\16\3\17")
        buf.write("\3\17\3\17\3\17\7\17\u00a8\n\17\f\17\16\17\u00ab\13\17")
        buf.write("\3\20\3\20\3\21\3\21\5\21\u00b1\n\21\3\22\3\22\3\22\3")
        buf.write("\22\3\22\3\22\3\22\3\22\3\22\5\22\u00bc\n\22\3\23\3\23")
        buf.write("\3\23\3\23\3\23\3\23\3\23\3\23\5\23\u00c6\n\23\3\24\3")
        buf.write("\24\3\24\3\24\3\24\7\24\u00cd\n\24\f\24\16\24\u00d0\13")
        buf.write("\24\3\24\3\24\3\25\3\25\7\25\u00d6\n\25\f\25\16\25\u00d9")
        buf.write("\13\25\3\26\3\26\3\27\3\27\3\27\3\27\3\27\5\27\u00e2\n")
        buf.write("\27\3\30\3\30\3\31\3\31\3\31\3\32\3\32\3\32\3\33\6\33")
        buf.write("\u00ed\n\33\r\33\16\33\u00ee\3\33\2\3\b\34\2\4\6\b\n\f")
        buf.write("\16\20\22\24\26\30\32\34\36 \"$&(*,.\60\62\64\2\5\3\2")
        buf.write("\n\17\3\2\20\21\3\2\34\35\2\u00f0\2\66\3\2\2\2\49\3\2")
        buf.write("\2\2\6A\3\2\2\2\bd\3\2\2\2\nq\3\2\2\2\fs\3\2\2\2\16z\3")
        buf.write("\2\2\2\20~\3\2\2\2\22\u0087\3\2\2\2\24\u008b\3\2\2\2\26")
        buf.write("\u0094\3\2\2\2\30\u0097\3\2\2\2\32\u00a0\3\2\2\2\34\u00a3")
        buf.write("\3\2\2\2\36\u00ac\3\2\2\2 \u00b0\3\2\2\2\"\u00bb\3\2\2")
        buf.write("\2$\u00c5\3\2\2\2&\u00c7\3\2\2\2(\u00d3\3\2\2\2*\u00da")
        buf.write("\3\2\2\2,\u00e1\3\2\2\2.\u00e3\3\2\2\2\60\u00e5\3\2\2")
        buf.write("\2\62\u00e8\3\2\2\2\64\u00ec\3\2\2\2\66\67\5\4\3\2\67")
        buf.write("8\7\2\2\38\3\3\2\2\29>\5(\25\2:;\7\3\2\2;=\5(\25\2<:\3")
        buf.write("\2\2\2=@\3\2\2\2><\3\2\2\2>?\3\2\2\2?\5\3\2\2\2@>\3\2")
        buf.write("\2\2AB\5\b\5\2BC\7\2\2\3C\7\3\2\2\2DE\b\5\1\2EF\5$\23")
        buf.write("\2FG\5\n\6\2GH\5$\23\2He\3\2\2\2IJ\5$\23\2JK\7\22\2\2")
        buf.write("KL\7\4\2\2LQ\5$\23\2MN\7\3\2\2NP\5$\23\2OM\3\2\2\2PS\3")
        buf.write("\2\2\2QO\3\2\2\2QR\3\2\2\2RT\3\2\2\2SQ\3\2\2\2TU\7\5\2")
        buf.write("\2Ue\3\2\2\2VW\5$\23\2WX\7\23\2\2XY\5$\23\2YZ\7\25\2\2")
        buf.write("Z[\5$\23\2[e\3\2\2\2\\e\5&\24\2]^\7\4\2\2^_\5\b\5\2_`")
        buf.write("\7\5\2\2`a\b\5\1\2ae\3\2\2\2bc\7\24\2\2ce\5\b\5\5dD\3")
        buf.write("\2\2\2dI\3\2\2\2dV\3\2\2\2d\\\3\2\2\2d]\3\2\2\2db\3\2")
        buf.write("\2\2en\3\2\2\2fg\f\4\2\2gh\7\25\2\2hm\5\b\5\5ij\f\3\2")
        buf.write("\2jk\7\26\2\2km\5\b\5\4lf\3\2\2\2li\3\2\2\2mp\3\2\2\2")
        buf.write("nl\3\2\2\2no\3\2\2\2o\t\3\2\2\2pn\3\2\2\2qr\t\2\2\2r\13")
        buf.write("\3\2\2\2st\5\16\b\2tu\7\2\2\3u\r\3\2\2\2v{\5\20\t\2w{")
        buf.write("\5\24\13\2x{\5\30\r\2y{\5\34\17\2zv\3\2\2\2zw\3\2\2\2")
        buf.write("zx\3\2\2\2zy\3\2\2\2{|\3\2\2\2|z\3\2\2\2|}\3\2\2\2}\17")
        buf.write("\3\2\2\2~\177\7\27\2\2\177\u0084\5\22\n\2\u0080\u0081")
        buf.write("\7\3\2\2\u0081\u0083\5\22\n\2\u0082\u0080\3\2\2\2\u0083")
        buf.write("\u0086\3\2\2\2\u0084\u0082\3\2\2\2\u0084\u0085\3\2\2\2")
        buf.write("\u0085\21\3\2\2\2\u0086\u0084\3\2\2\2\u0087\u0088\5(\25")
        buf.write("\2\u0088\u0089\7\n\2\2\u0089\u008a\5 \21\2\u008a\23\3")
        buf.write("\2\2\2\u008b\u008c\7\30\2\2\u008c\u0091\5\26\f\2\u008d")
        buf.write("\u008e\7\3\2\2\u008e\u0090\5\26\f\2\u008f\u008d\3\2\2")
        buf.write("\2\u0090\u0093\3\2\2\2\u0091\u008f\3\2\2\2\u0091\u0092")
        buf.write("\3\2\2\2\u0092\25\3\2\2\2\u0093\u0091\3\2\2\2\u0094\u0095")
        buf.write("\5(\25\2\u0095\u0096\5.\30\2\u0096\27\3\2\2\2\u0097\u0098")
        buf.write("\7\31\2\2\u0098\u009d\5\32\16\2\u0099\u009a\7\3\2\2\u009a")
        buf.write("\u009c\5\32\16\2\u009b\u0099\3\2\2\2\u009c\u009f\3\2\2")
        buf.write("\2\u009d\u009b\3\2\2\2\u009d\u009e\3\2\2\2\u009e\31\3")
        buf.write("\2\2\2\u009f\u009d\3\2\2\2\u00a0\u00a1\5(\25\2\u00a1\u00a2")
        buf.write("\5.\30\2\u00a2\33\3\2\2\2\u00a3\u00a4\7\32\2\2\u00a4\u00a9")
        buf.write("\5\36\20\2\u00a5\u00a6\7\3\2\2\u00a6\u00a8\5\36\20\2\u00a7")
        buf.write("\u00a5\3\2\2\2\u00a8\u00ab\3\2\2\2\u00a9\u00a7\3\2\2\2")
        buf.write("\u00a9\u00aa\3\2\2\2\u00aa\35\3\2\2\2\u00ab\u00a9\3\2")
        buf.write("\2\2\u00ac\u00ad\5(\25\2\u00ad\37\3\2\2\2\u00ae\u00b1")
        buf.write("\5$\23\2\u00af\u00b1\5\"\22\2\u00b0\u00ae\3\2\2\2\u00b0")
        buf.write("\u00af\3\2\2\2\u00b1!\3\2\2\2\u00b2\u00b3\5$\23\2\u00b3")
        buf.write("\u00b4\t\3\2\2\u00b4\u00b5\5$\23\2\u00b5\u00bc\3\2\2\2")
        buf.write("\u00b6\u00b7\7\4\2\2\u00b7\u00b8\5\"\22\2\u00b8\u00b9")
        buf.write("\7\5\2\2\u00b9\u00ba\b\22\1\2\u00ba\u00bc\3\2\2\2\u00bb")
        buf.write("\u00b2\3\2\2\2\u00bb\u00b6\3\2\2\2\u00bc#\3\2\2\2\u00bd")
        buf.write("\u00c6\5(\25\2\u00be\u00c6\5.\30\2\u00bf\u00c6\5&\24\2")
        buf.write("\u00c0\u00c1\7\4\2\2\u00c1\u00c2\5$\23\2\u00c2\u00c3\7")
        buf.write("\5\2\2\u00c3\u00c4\b\23\1\2\u00c4\u00c6\3\2\2\2\u00c5")
        buf.write("\u00bd\3\2\2\2\u00c5\u00be\3\2\2\2\u00c5\u00bf\3\2\2\2")
        buf.write("\u00c5\u00c0\3\2\2\2\u00c6%\3\2\2\2\u00c7\u00c8\7\34\2")
        buf.write("\2\u00c8\u00c9\7\4\2\2\u00c9\u00ce\5$\23\2\u00ca\u00cb")
        buf.write("\7\3\2\2\u00cb\u00cd\5$\23\2\u00cc\u00ca\3\2\2\2\u00cd")
        buf.write("\u00d0\3\2\2\2\u00ce\u00cc\3\2\2\2\u00ce\u00cf\3\2\2\2")
        buf.write("\u00cf\u00d1\3\2\2\2\u00d0\u00ce\3\2\2\2\u00d1\u00d2\7")
        buf.write("\5\2\2\u00d2\'\3\2\2\2\u00d3\u00d7\5*\26\2\u00d4\u00d6")
        buf.write("\5,\27\2\u00d5\u00d4\3\2\2\2\u00d6\u00d9\3\2\2\2\u00d7")
        buf.write("\u00d5\3\2\2\2\u00d7\u00d8\3\2\2\2\u00d8)\3\2\2\2\u00d9")
        buf.write("\u00d7\3\2\2\2\u00da\u00db\t\4\2\2\u00db+\3\2\2\2\u00dc")
        buf.write("\u00dd\7\6\2\2\u00dd\u00e2\5*\26\2\u00de\u00df\7\7\2\2")
        buf.write("\u00df\u00e0\7\33\2\2\u00e0\u00e2\7\b\2\2\u00e1\u00dc")
        buf.write("\3\2\2\2\u00e1\u00de\3\2\2\2\u00e2-\3\2\2\2\u00e3\u00e4")
        buf.write("\7\36\2\2\u00e4/\3\2\2\2\u00e5\u00e6\7\35\2\2\u00e6\u00e7")
        buf.write("\7\2\2\3\u00e7\61\3\2\2\2\u00e8\u00e9\7\36\2\2\u00e9\u00ea")
        buf.write("\7\2\2\3\u00ea\63\3\2\2\2\u00eb\u00ed\7 \2\2\u00ec\u00eb")
        buf.write("\3\2\2\2\u00ed\u00ee\3\2\2\2\u00ee\u00ec\3\2\2\2\u00ee")
        buf.write("\u00ef\3\2\2\2\u00ef\65\3\2\2\2\24>Qdlnz|\u0084\u0091")
        buf.write("\u009d\u00a9\u00b0\u00bb\u00c5\u00ce\u00d7\u00e1\u00ee")
        return buf.getvalue()


class DynamoDbGrammarParser ( Parser ):

    grammarFileName = "DynamoDbGrammar.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "','", "'('", "')'", "'.'", "'['", "']'",
                     "<INVALID>", "'='", "'<>'", "'<'", "'<='", "'>'", "'>='",
                     "'+'", "'-'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                      "<INVALID>", "<INVALID>", "<INVALID>", "WS", "EQ",
                      "NE", "LT", "LE", "GT", "GE", "PLUS", "MINUS", "IN",
                      "BETWEEN", "NOT", "AND", "OR", "SET", "ADD", "DELETE",
                      "REMOVE", "INDEX", "ID", "ATTRIBUTE_NAME_SUB", "LITERAL_SUB",
                      "STRING_LITERAL", "UNKNOWN" ]

    RULE_projection_ = 0
    RULE_projection = 1
    RULE_condition_ = 2
    RULE_condition = 3
    RULE_comparator_symbol = 4
    RULE_update_ = 5
    RULE_update = 6
    RULE_set_section = 7
    RULE_set_action = 8
    RULE_add_section = 9
    RULE_add_action = 10
    RULE_delete_section = 11
    RULE_delete_action = 12
    RULE_remove_section = 13
    RULE_remove_action = 14
    RULE_set_value = 15
    RULE_arithmetic = 16
    RULE_operand = 17
    RULE_func = 18
    RULE_path = 19
    RULE_id_ = 20
    RULE_dereference = 21
    RULE_literal = 22
    RULE_expression_attr_names_sub = 23
    RULE_expression_attr_values_sub = 24
    RULE_unknown = 25

    ruleNames =  [ "projection_", "projection", "condition_", "condition",
                   "comparator_symbol", "update_", "update", "set_section",
                   "set_action", "add_section", "add_action", "delete_section",
                   "delete_action", "remove_section", "remove_action", "set_value",
                   "arithmetic", "operand", "func", "path", "id_", "dereference",
                   "literal", "expression_attr_names_sub", "expression_attr_values_sub",
                   "unknown" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    WS=7
    EQ=8
    NE=9
    LT=10
    LE=11
    GT=12
    GE=13
    PLUS=14
    MINUS=15
    IN=16
    BETWEEN=17
    NOT=18
    AND=19
    OR=20
    SET=21
    ADD=22
    DELETE=23
    REMOVE=24
    INDEX=25
    ID=26
    ATTRIBUTE_NAME_SUB=27
    LITERAL_SUB=28
    STRING_LITERAL=29
    UNKNOWN=30

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    def validateRedundantParentheses(self, redundantParens):
        if redundantParens:
            raise Exception('RedundantParenthesesException')



    class Projection_Context(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def projection(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.ProjectionContext,0)


        def EOF(self):
            return self.getToken(DynamoDbGrammarParser.EOF, 0)

        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_projection_

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterProjection_" ):
                listener.enterProjection_(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitProjection_" ):
                listener.exitProjection_(self)




    def projection_(self):

        localctx = DynamoDbGrammarParser.Projection_Context(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_projection_)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 52
            self.projection()
            self.state = 53
            self.match(DynamoDbGrammarParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ProjectionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def path(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.PathContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.PathContext,i)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_projection

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterProjection" ):
                listener.enterProjection(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitProjection" ):
                listener.exitProjection(self)




    def projection(self):

        localctx = DynamoDbGrammarParser.ProjectionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_projection)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 55
            self.path()
            self.state = 60
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==DynamoDbGrammarParser.T__0:
                self.state = 56
                self.match(DynamoDbGrammarParser.T__0)
                self.state = 57
                self.path()
                self.state = 62
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Condition_Context(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def condition(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.ConditionContext,0)


        def EOF(self):
            return self.getToken(DynamoDbGrammarParser.EOF, 0)

        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_condition_

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCondition_" ):
                listener.enterCondition_(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCondition_" ):
                listener.exitCondition_(self)




    def condition_(self):

        localctx = DynamoDbGrammarParser.Condition_Context(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_condition_)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 63
            self.condition(0)
            self.state = 64
            self.match(DynamoDbGrammarParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ConditionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.hasOuterParens = False


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_condition


        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)
            self.hasOuterParens = ctx.hasOuterParens


    class OrContext(ConditionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def condition(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.ConditionContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.ConditionContext,i)

        def OR(self):
            return self.getToken(DynamoDbGrammarParser.OR, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOr" ):
                listener.enterOr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOr" ):
                listener.exitOr(self)


    class NegationContext(ConditionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NOT(self):
            return self.getToken(DynamoDbGrammarParser.NOT, 0)
        def condition(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.ConditionContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNegation" ):
                listener.enterNegation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNegation" ):
                listener.exitNegation(self)


    class InContext(ConditionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def operand(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.OperandContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.OperandContext,i)

        def IN(self):
            return self.getToken(DynamoDbGrammarParser.IN, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterIn" ):
                listener.enterIn(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitIn" ):
                listener.exitIn(self)


    class AndContext(ConditionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def condition(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.ConditionContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.ConditionContext,i)

        def AND(self):
            return self.getToken(DynamoDbGrammarParser.AND, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAnd" ):
                listener.enterAnd(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAnd" ):
                listener.exitAnd(self)


    class BetweenContext(ConditionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def operand(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.OperandContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.OperandContext,i)

        def BETWEEN(self):
            return self.getToken(DynamoDbGrammarParser.BETWEEN, 0)
        def AND(self):
            return self.getToken(DynamoDbGrammarParser.AND, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBetween" ):
                listener.enterBetween(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBetween" ):
                listener.exitBetween(self)


    class FunctionConditionContext(ConditionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def func(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.FuncContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFunctionCondition" ):
                listener.enterFunctionCondition(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFunctionCondition" ):
                listener.exitFunctionCondition(self)


    class ComparatorContext(ConditionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def operand(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.OperandContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.OperandContext,i)

        def comparator_symbol(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.Comparator_symbolContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComparator" ):
                listener.enterComparator(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComparator" ):
                listener.exitComparator(self)


    class ConditionGroupingContext(ConditionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.ConditionContext
            super().__init__(parser)
            self.c = None # ConditionContext
            self.copyFrom(ctx)

        def condition(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.ConditionContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterConditionGrouping" ):
                listener.enterConditionGrouping(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitConditionGrouping" ):
                listener.exitConditionGrouping(self)



    def condition(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = DynamoDbGrammarParser.ConditionContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 6
        self.enterRecursionRule(localctx, 6, self.RULE_condition, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 98
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,2,self._ctx)
            if la_ == 1:
                localctx = DynamoDbGrammarParser.ComparatorContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 67
                self.operand()
                self.state = 68
                self.comparator_symbol()
                self.state = 69
                self.operand()
                pass

            elif la_ == 2:
                localctx = DynamoDbGrammarParser.InContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 71
                self.operand()
                self.state = 72
                self.match(DynamoDbGrammarParser.IN)
                self.state = 73
                self.match(DynamoDbGrammarParser.T__1)
                self.state = 74
                self.operand()
                self.state = 79
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==DynamoDbGrammarParser.T__0:
                    self.state = 75
                    self.match(DynamoDbGrammarParser.T__0)
                    self.state = 76
                    self.operand()
                    self.state = 81
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 82
                self.match(DynamoDbGrammarParser.T__2)
                pass

            elif la_ == 3:
                localctx = DynamoDbGrammarParser.BetweenContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 84
                self.operand()
                self.state = 85
                self.match(DynamoDbGrammarParser.BETWEEN)
                self.state = 86
                self.operand()
                self.state = 87
                self.match(DynamoDbGrammarParser.AND)
                self.state = 88
                self.operand()
                pass

            elif la_ == 4:
                localctx = DynamoDbGrammarParser.FunctionConditionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 90
                self.func()
                pass

            elif la_ == 5:
                localctx = DynamoDbGrammarParser.ConditionGroupingContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 91
                self.match(DynamoDbGrammarParser.T__1)
                self.state = 92
                localctx.c = self.condition(0)
                self.state = 93
                self.match(DynamoDbGrammarParser.T__2)

                self.validateRedundantParentheses(localctx.c.hasOuterParens)
                localctx.hasOuterParens=True

                pass

            elif la_ == 6:
                localctx = DynamoDbGrammarParser.NegationContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 96
                self.match(DynamoDbGrammarParser.NOT)
                self.state = 97
                self.condition(3)
                pass


            self._ctx.stop = self._input.LT(-1)
            self.state = 108
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,4,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 106
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,3,self._ctx)
                    if la_ == 1:
                        localctx = DynamoDbGrammarParser.AndContext(self, DynamoDbGrammarParser.ConditionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_condition)
                        self.state = 100
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 101
                        self.match(DynamoDbGrammarParser.AND)
                        self.state = 102
                        self.condition(3)
                        pass

                    elif la_ == 2:
                        localctx = DynamoDbGrammarParser.OrContext(self, DynamoDbGrammarParser.ConditionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_condition)
                        self.state = 103
                        if not self.precpred(self._ctx, 1):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 1)")
                        self.state = 104
                        self.match(DynamoDbGrammarParser.OR)
                        self.state = 105
                        self.condition(2)
                        pass


                self.state = 110
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,4,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx


    class Comparator_symbolContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EQ(self):
            return self.getToken(DynamoDbGrammarParser.EQ, 0)

        def NE(self):
            return self.getToken(DynamoDbGrammarParser.NE, 0)

        def LT(self):
            return self.getToken(DynamoDbGrammarParser.LT, 0)

        def LE(self):
            return self.getToken(DynamoDbGrammarParser.LE, 0)

        def GT(self):
            return self.getToken(DynamoDbGrammarParser.GT, 0)

        def GE(self):
            return self.getToken(DynamoDbGrammarParser.GE, 0)

        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_comparator_symbol

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComparator_symbol" ):
                listener.enterComparator_symbol(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComparator_symbol" ):
                listener.exitComparator_symbol(self)




    def comparator_symbol(self):

        localctx = DynamoDbGrammarParser.Comparator_symbolContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_comparator_symbol)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 111
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << DynamoDbGrammarParser.EQ) | (1 << DynamoDbGrammarParser.NE) | (1 << DynamoDbGrammarParser.LT) | (1 << DynamoDbGrammarParser.LE) | (1 << DynamoDbGrammarParser.GT) | (1 << DynamoDbGrammarParser.GE))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Update_Context(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def update(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.UpdateContext,0)


        def EOF(self):
            return self.getToken(DynamoDbGrammarParser.EOF, 0)

        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_update_

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterUpdate_" ):
                listener.enterUpdate_(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitUpdate_" ):
                listener.exitUpdate_(self)




    def update_(self):

        localctx = DynamoDbGrammarParser.Update_Context(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_update_)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 113
            self.update()
            self.state = 114
            self.match(DynamoDbGrammarParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class UpdateContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def set_section(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.Set_sectionContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.Set_sectionContext,i)


        def add_section(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.Add_sectionContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.Add_sectionContext,i)


        def delete_section(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.Delete_sectionContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.Delete_sectionContext,i)


        def remove_section(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.Remove_sectionContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.Remove_sectionContext,i)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_update

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterUpdate" ):
                listener.enterUpdate(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitUpdate" ):
                listener.exitUpdate(self)




    def update(self):

        localctx = DynamoDbGrammarParser.UpdateContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_update)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 120
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 120
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [DynamoDbGrammarParser.SET]:
                    self.state = 116
                    self.set_section()
                    pass
                elif token in [DynamoDbGrammarParser.ADD]:
                    self.state = 117
                    self.add_section()
                    pass
                elif token in [DynamoDbGrammarParser.DELETE]:
                    self.state = 118
                    self.delete_section()
                    pass
                elif token in [DynamoDbGrammarParser.REMOVE]:
                    self.state = 119
                    self.remove_section()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 122
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << DynamoDbGrammarParser.SET) | (1 << DynamoDbGrammarParser.ADD) | (1 << DynamoDbGrammarParser.DELETE) | (1 << DynamoDbGrammarParser.REMOVE))) != 0)):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Set_sectionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def SET(self):
            return self.getToken(DynamoDbGrammarParser.SET, 0)

        def set_action(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.Set_actionContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.Set_actionContext,i)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_set_section

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSet_section" ):
                listener.enterSet_section(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSet_section" ):
                listener.exitSet_section(self)




    def set_section(self):

        localctx = DynamoDbGrammarParser.Set_sectionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_set_section)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 124
            self.match(DynamoDbGrammarParser.SET)
            self.state = 125
            self.set_action()
            self.state = 130
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==DynamoDbGrammarParser.T__0:
                self.state = 126
                self.match(DynamoDbGrammarParser.T__0)
                self.state = 127
                self.set_action()
                self.state = 132
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Set_actionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def path(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.PathContext,0)


        def EQ(self):
            return self.getToken(DynamoDbGrammarParser.EQ, 0)

        def set_value(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.Set_valueContext,0)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_set_action

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSet_action" ):
                listener.enterSet_action(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSet_action" ):
                listener.exitSet_action(self)




    def set_action(self):

        localctx = DynamoDbGrammarParser.Set_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_set_action)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 133
            self.path()
            self.state = 134
            self.match(DynamoDbGrammarParser.EQ)
            self.state = 135
            self.set_value()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Add_sectionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ADD(self):
            return self.getToken(DynamoDbGrammarParser.ADD, 0)

        def add_action(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.Add_actionContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.Add_actionContext,i)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_add_section

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAdd_section" ):
                listener.enterAdd_section(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAdd_section" ):
                listener.exitAdd_section(self)




    def add_section(self):

        localctx = DynamoDbGrammarParser.Add_sectionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_add_section)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 137
            self.match(DynamoDbGrammarParser.ADD)
            self.state = 138
            self.add_action()
            self.state = 143
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==DynamoDbGrammarParser.T__0:
                self.state = 139
                self.match(DynamoDbGrammarParser.T__0)
                self.state = 140
                self.add_action()
                self.state = 145
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Add_actionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def path(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.PathContext,0)


        def literal(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.LiteralContext,0)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_add_action

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAdd_action" ):
                listener.enterAdd_action(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAdd_action" ):
                listener.exitAdd_action(self)




    def add_action(self):

        localctx = DynamoDbGrammarParser.Add_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_add_action)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 146
            self.path()
            self.state = 147
            self.literal()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Delete_sectionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DELETE(self):
            return self.getToken(DynamoDbGrammarParser.DELETE, 0)

        def delete_action(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.Delete_actionContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.Delete_actionContext,i)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_delete_section

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDelete_section" ):
                listener.enterDelete_section(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDelete_section" ):
                listener.exitDelete_section(self)




    def delete_section(self):

        localctx = DynamoDbGrammarParser.Delete_sectionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_delete_section)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 149
            self.match(DynamoDbGrammarParser.DELETE)
            self.state = 150
            self.delete_action()
            self.state = 155
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==DynamoDbGrammarParser.T__0:
                self.state = 151
                self.match(DynamoDbGrammarParser.T__0)
                self.state = 152
                self.delete_action()
                self.state = 157
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Delete_actionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def path(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.PathContext,0)


        def literal(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.LiteralContext,0)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_delete_action

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDelete_action" ):
                listener.enterDelete_action(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDelete_action" ):
                listener.exitDelete_action(self)




    def delete_action(self):

        localctx = DynamoDbGrammarParser.Delete_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_delete_action)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 158
            self.path()
            self.state = 159
            self.literal()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Remove_sectionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def REMOVE(self):
            return self.getToken(DynamoDbGrammarParser.REMOVE, 0)

        def remove_action(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.Remove_actionContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.Remove_actionContext,i)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_remove_section

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRemove_section" ):
                listener.enterRemove_section(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRemove_section" ):
                listener.exitRemove_section(self)




    def remove_section(self):

        localctx = DynamoDbGrammarParser.Remove_sectionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_remove_section)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 161
            self.match(DynamoDbGrammarParser.REMOVE)
            self.state = 162
            self.remove_action()
            self.state = 167
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==DynamoDbGrammarParser.T__0:
                self.state = 163
                self.match(DynamoDbGrammarParser.T__0)
                self.state = 164
                self.remove_action()
                self.state = 169
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Remove_actionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def path(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.PathContext,0)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_remove_action

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRemove_action" ):
                listener.enterRemove_action(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRemove_action" ):
                listener.exitRemove_action(self)




    def remove_action(self):

        localctx = DynamoDbGrammarParser.Remove_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_remove_action)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 170
            self.path()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Set_valueContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_set_value


        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class ArithmeticValueContext(Set_valueContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.Set_valueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def arithmetic(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.ArithmeticContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterArithmeticValue" ):
                listener.enterArithmeticValue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitArithmeticValue" ):
                listener.exitArithmeticValue(self)


    class OperandValueContext(Set_valueContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.Set_valueContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def operand(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.OperandContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOperandValue" ):
                listener.enterOperandValue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOperandValue" ):
                listener.exitOperandValue(self)



    def set_value(self):

        localctx = DynamoDbGrammarParser.Set_valueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_set_value)
        try:
            self.state = 174
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,11,self._ctx)
            if la_ == 1:
                localctx = DynamoDbGrammarParser.OperandValueContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 172
                self.operand()
                pass

            elif la_ == 2:
                localctx = DynamoDbGrammarParser.ArithmeticValueContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 173
                self.arithmetic()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ArithmeticContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.hasOuterParens = False


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_arithmetic


        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)
            self.hasOuterParens = ctx.hasOuterParens



    class PlusMinusContext(ArithmeticContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.ArithmeticContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def operand(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.OperandContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.OperandContext,i)

        def PLUS(self):
            return self.getToken(DynamoDbGrammarParser.PLUS, 0)
        def MINUS(self):
            return self.getToken(DynamoDbGrammarParser.MINUS, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPlusMinus" ):
                listener.enterPlusMinus(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPlusMinus" ):
                listener.exitPlusMinus(self)


    class ArithmeticParensContext(ArithmeticContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.ArithmeticContext
            super().__init__(parser)
            self.a = None # ArithmeticContext
            self.copyFrom(ctx)

        def arithmetic(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.ArithmeticContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterArithmeticParens" ):
                listener.enterArithmeticParens(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitArithmeticParens" ):
                listener.exitArithmeticParens(self)



    def arithmetic(self):

        localctx = DynamoDbGrammarParser.ArithmeticContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_arithmetic)
        self._la = 0 # Token type
        try:
            self.state = 185
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,12,self._ctx)
            if la_ == 1:
                localctx = DynamoDbGrammarParser.PlusMinusContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 176
                self.operand()
                self.state = 177
                _la = self._input.LA(1)
                if not(_la==DynamoDbGrammarParser.PLUS or _la==DynamoDbGrammarParser.MINUS):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 178
                self.operand()
                pass

            elif la_ == 2:
                localctx = DynamoDbGrammarParser.ArithmeticParensContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 180
                self.match(DynamoDbGrammarParser.T__1)
                self.state = 181
                localctx.a = self.arithmetic()
                self.state = 182
                self.match(DynamoDbGrammarParser.T__2)

                self.validateRedundantParentheses(localctx.a.hasOuterParens)
                localctx.hasOuterParens=True

                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class OperandContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.hasOuterParens = False


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_operand


        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)
            self.hasOuterParens = ctx.hasOuterParens



    class PathOperandContext(OperandContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.OperandContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def path(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.PathContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPathOperand" ):
                listener.enterPathOperand(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPathOperand" ):
                listener.exitPathOperand(self)


    class LiteralOperandContext(OperandContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.OperandContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def literal(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.LiteralContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLiteralOperand" ):
                listener.enterLiteralOperand(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLiteralOperand" ):
                listener.exitLiteralOperand(self)


    class FunctionOperandContext(OperandContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.OperandContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def func(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.FuncContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFunctionOperand" ):
                listener.enterFunctionOperand(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFunctionOperand" ):
                listener.exitFunctionOperand(self)


    class ParenOperandContext(OperandContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.OperandContext
            super().__init__(parser)
            self.o = None # OperandContext
            self.copyFrom(ctx)

        def operand(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.OperandContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterParenOperand" ):
                listener.enterParenOperand(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitParenOperand" ):
                listener.exitParenOperand(self)



    def operand(self):

        localctx = DynamoDbGrammarParser.OperandContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_operand)
        try:
            self.state = 195
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,13,self._ctx)
            if la_ == 1:
                localctx = DynamoDbGrammarParser.PathOperandContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 187
                self.path()
                pass

            elif la_ == 2:
                localctx = DynamoDbGrammarParser.LiteralOperandContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 188
                self.literal()
                pass

            elif la_ == 3:
                localctx = DynamoDbGrammarParser.FunctionOperandContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 189
                self.func()
                pass

            elif la_ == 4:
                localctx = DynamoDbGrammarParser.ParenOperandContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 190
                self.match(DynamoDbGrammarParser.T__1)
                self.state = 191
                localctx.o = self.operand()
                self.state = 192
                self.match(DynamoDbGrammarParser.T__2)

                self.validateRedundantParentheses(localctx.o.hasOuterParens)
                localctx.hasOuterParens=True

                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class FuncContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_func


        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class FunctionCallContext(FuncContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.FuncContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(DynamoDbGrammarParser.ID, 0)
        def operand(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.OperandContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.OperandContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFunctionCall" ):
                listener.enterFunctionCall(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFunctionCall" ):
                listener.exitFunctionCall(self)



    def func(self):

        localctx = DynamoDbGrammarParser.FuncContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_func)
        self._la = 0 # Token type
        try:
            localctx = DynamoDbGrammarParser.FunctionCallContext(self, localctx)
            self.enterOuterAlt(localctx, 1)
            self.state = 197
            self.match(DynamoDbGrammarParser.ID)
            self.state = 198
            self.match(DynamoDbGrammarParser.T__1)
            self.state = 199
            self.operand()
            self.state = 204
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==DynamoDbGrammarParser.T__0:
                self.state = 200
                self.match(DynamoDbGrammarParser.T__0)
                self.state = 201
                self.operand()
                self.state = 206
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 207
            self.match(DynamoDbGrammarParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PathContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def id_(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.Id_Context,0)


        def dereference(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynamoDbGrammarParser.DereferenceContext)
            else:
                return self.getTypedRuleContext(DynamoDbGrammarParser.DereferenceContext,i)


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_path

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPath" ):
                listener.enterPath(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPath" ):
                listener.exitPath(self)




    def path(self):

        localctx = DynamoDbGrammarParser.PathContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_path)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 209
            self.id_()
            self.state = 213
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,15,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 210
                    self.dereference()
                self.state = 215
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,15,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Id_Context(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(DynamoDbGrammarParser.ID, 0)

        def ATTRIBUTE_NAME_SUB(self):
            return self.getToken(DynamoDbGrammarParser.ATTRIBUTE_NAME_SUB, 0)

        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_id_

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterId_" ):
                listener.enterId_(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitId_" ):
                listener.exitId_(self)




    def id_(self):

        localctx = DynamoDbGrammarParser.Id_Context(self, self._ctx, self.state)
        self.enterRule(localctx, 40, self.RULE_id_)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 216
            _la = self._input.LA(1)
            if not(_la==DynamoDbGrammarParser.ID or _la==DynamoDbGrammarParser.ATTRIBUTE_NAME_SUB):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DereferenceContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_dereference


        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class ListAccessContext(DereferenceContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.DereferenceContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def INDEX(self):
            return self.getToken(DynamoDbGrammarParser.INDEX, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterListAccess" ):
                listener.enterListAccess(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitListAccess" ):
                listener.exitListAccess(self)


    class MapAccessContext(DereferenceContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.DereferenceContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def id_(self):
            return self.getTypedRuleContext(DynamoDbGrammarParser.Id_Context,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMapAccess" ):
                listener.enterMapAccess(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMapAccess" ):
                listener.exitMapAccess(self)



    def dereference(self):

        localctx = DynamoDbGrammarParser.DereferenceContext(self, self._ctx, self.state)
        self.enterRule(localctx, 42, self.RULE_dereference)
        try:
            self.state = 223
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [DynamoDbGrammarParser.T__3]:
                localctx = DynamoDbGrammarParser.MapAccessContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 218
                self.match(DynamoDbGrammarParser.T__3)
                self.state = 219
                self.id_()
                pass
            elif token in [DynamoDbGrammarParser.T__4]:
                localctx = DynamoDbGrammarParser.ListAccessContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 220
                self.match(DynamoDbGrammarParser.T__4)
                self.state = 221
                self.match(DynamoDbGrammarParser.INDEX)
                self.state = 222
                self.match(DynamoDbGrammarParser.T__5)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class LiteralContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_literal


        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class LiteralSubContext(LiteralContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DynamoDbGrammarParser.LiteralContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LITERAL_SUB(self):
            return self.getToken(DynamoDbGrammarParser.LITERAL_SUB, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLiteralSub" ):
                listener.enterLiteralSub(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLiteralSub" ):
                listener.exitLiteralSub(self)



    def literal(self):

        localctx = DynamoDbGrammarParser.LiteralContext(self, self._ctx, self.state)
        self.enterRule(localctx, 44, self.RULE_literal)
        try:
            localctx = DynamoDbGrammarParser.LiteralSubContext(self, localctx)
            self.enterOuterAlt(localctx, 1)
            self.state = 225
            self.match(DynamoDbGrammarParser.LITERAL_SUB)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Expression_attr_names_subContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ATTRIBUTE_NAME_SUB(self):
            return self.getToken(DynamoDbGrammarParser.ATTRIBUTE_NAME_SUB, 0)

        def EOF(self):
            return self.getToken(DynamoDbGrammarParser.EOF, 0)

        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_expression_attr_names_sub

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExpression_attr_names_sub" ):
                listener.enterExpression_attr_names_sub(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExpression_attr_names_sub" ):
                listener.exitExpression_attr_names_sub(self)




    def expression_attr_names_sub(self):

        localctx = DynamoDbGrammarParser.Expression_attr_names_subContext(self, self._ctx, self.state)
        self.enterRule(localctx, 46, self.RULE_expression_attr_names_sub)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 227
            self.match(DynamoDbGrammarParser.ATTRIBUTE_NAME_SUB)
            self.state = 228
            self.match(DynamoDbGrammarParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Expression_attr_values_subContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LITERAL_SUB(self):
            return self.getToken(DynamoDbGrammarParser.LITERAL_SUB, 0)

        def EOF(self):
            return self.getToken(DynamoDbGrammarParser.EOF, 0)

        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_expression_attr_values_sub

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExpression_attr_values_sub" ):
                listener.enterExpression_attr_values_sub(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExpression_attr_values_sub" ):
                listener.exitExpression_attr_values_sub(self)




    def expression_attr_values_sub(self):

        localctx = DynamoDbGrammarParser.Expression_attr_values_subContext(self, self._ctx, self.state)
        self.enterRule(localctx, 48, self.RULE_expression_attr_values_sub)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 230
            self.match(DynamoDbGrammarParser.LITERAL_SUB)
            self.state = 231
            self.match(DynamoDbGrammarParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class UnknownContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def UNKNOWN(self, i:int=None):
            if i is None:
                return self.getTokens(DynamoDbGrammarParser.UNKNOWN)
            else:
                return self.getToken(DynamoDbGrammarParser.UNKNOWN, i)

        def getRuleIndex(self):
            return DynamoDbGrammarParser.RULE_unknown

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterUnknown" ):
                listener.enterUnknown(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitUnknown" ):
                listener.exitUnknown(self)




    def unknown(self):

        localctx = DynamoDbGrammarParser.UnknownContext(self, self._ctx, self.state)
        self.enterRule(localctx, 50, self.RULE_unknown)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 234
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 233
                self.match(DynamoDbGrammarParser.UNKNOWN)
                self.state = 236
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==DynamoDbGrammarParser.UNKNOWN):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



    def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[3] = self.condition_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def condition_sempred(self, localctx:ConditionContext, predIndex:int):
            if predIndex == 0:
                return self.precpred(self._ctx, 2)


            if predIndex == 1:
                return self.precpred(self._ctx, 1)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import six

if six.PY2:
    from amazondax.grammar2.DynamoDbGrammarLexer import DynamoDbGrammarLexer
    from amazondax.grammar2.DynamoDbGrammarParser import DynamoDbGrammarParser
else:
    from amazondax.grammar.DynamoDbGrammarLexer import DynamoDbGrammarLexer # pylint: disable=import-error,no-name-in-module
    from amazondax.grammar.DynamoDbGrammarParser import DynamoDbGrammarParser # pylint: disable=import-error,no-name-in-module

import antlr4
from antlr4.error.ErrorStrategy import DefaultErrorStrategy, BailErrorStrategy
from antlr4.error.Errors import ParseCancellationException

def parse_projection(expression, error_listener):
    return _ProjectionExpression().parse(expression, error_listener)

def parse_condition(expression, error_listener):
    return _ConditionExpression().parse(expression, error_listener)

def parse_update(expression, error_listener):
    return _UpdateExpression().parse(expression, error_listener)

class _AbstractTwoStageExpressionParser(object):
    def parse(self, expression, error_listener):
        lexer = DynamoDbGrammarLexer(antlr4.InputStream(expression))
        tokens = antlr4.CommonTokenStream(lexer)
        parser = DynamoDbGrammarParser(tokens)
        parser.buildParseTrees = True

        lexer.removeErrorListeners()
        lexer.addErrorListener(error_listener)
        parser.removeErrorListeners()

        # DO NOT CALL addErrorListener(errorListener) for BailErrorStrategy
        # ExpressionErrorListener converts syntaxErrors to validation exceptions
        # But when using SLL strategy syntaxErrors can be false
        # Such that a syntaxError thrown by SLL may not necessarily be a real syntaxError
        # For such cases we don't want the syntaxError become a validation exception and go up to customers
        # We need to parse the expression again with LL strategy in case of a syntax error
        # BailErrorStrategy will re-throw RecognitionExceptions as ParseCancellationException
        # Such that it's not caught by underlying parsing rule implemented by parseStub
        parser._interp.predictionMode = antlr4.PredictionMode.SLL # pylint: disable=protected-access
        parser._errHandler = BailErrorStrategy() # pylint: disable=protected-access
        try:
            # Stage 1 parse with PredictionMode.SLL
            # If there are no issues SLL was enough to parse
            # If there were problems LL will be used to try again
            return self.parse_stub(parser)
        except ParseCancellationException:
            # If there was an error we don't know if it's a real SyntaxError
            # Or SLL strategy wasn't strong enough
            # Stage 2 parse with default prediction mode
            tokens.reset()
            parser.reset()
            parser.addErrorListener(error_listener)
            parser._errHandler = DefaultErrorStrategy() # pylint: disable=protected-access
            parser._interp.predictionMode = antlr4.PredictionMode.LL # pylint: disable=protected-access
            return self.parse_stub(parser)

    def parse_stub(self, parser):
        raise NotImplementedError('abstract method')

class _ProjectionExpression(_AbstractTwoStageExpressionParser):
    def parse_stub(self, parser):
        return parser.projection_()

class _ConditionExpression(_AbstractTwoStageExpressionParser):
    def parse_stub(self, parser):
        return parser.condition_()

class _UpdateExpression(_AbstractTwoStageExpressionParser):
    def parse_stub(self, parser):
        return parser.update_()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from . import DynamoDbExpressionParser, AttributeValueEncoder, DaxCborTypes
from .CborEncoder import CborEncoder
from .DaxError import DaxClientError, DaxErrorCode, DaxValidationError

import six

if six.PY2:
    from .grammar2.DynamoDbGrammarListener import DynamoDbGrammarListener
else:
    from .grammar.DynamoDbGrammarListener import DynamoDbGrammarListener # pylint: disable=import-error,no-name-in-module

import antlr4

# TODO I feel like this should be able to take a CborEncoder and fill it
# directly, rather than build a bunch of bytes for appending. But maybe
# in this case it's better to be consistent with other clients than
# clever.

class Expressions(object):
    def __init__(self):
        self.Condition = None
        self.KeyCondition = None
        self.Filter = None
        self.Update = None
        self.Projection = None
        self.ExpressionAttributeNames = None
        self.ExpressionAttributeValues = None

def encode_condition_expression(condition_expr, expr_attr_names, expr_attr_values):
    return encode_expressions(condition_expr, None, None, None, None, expr_attr_names, expr_attr_values).Condition

def encode_key_condition_expression(key_condition_expr, expr_attr_names, expr_attr_values):
    return encode_expressions(None, key_condition_expr, None, None, None, expr_attr_names, expr_attr_values).KeyCondition

def encode_filter_expression(filter_expr, expr_attr_names, expr_attr_values):
    return encode_expressions(None, None, filter_expr, None, None, expr_attr_names, expr_attr_values).Filter

def encode_update_expression(update_expr, expr_attr_names, expr_attr_values):
    return encode_expressions(None, None, None, update_expr, None, expr_attr_names, expr_attr_values).Update

def encode_projection_expression(projection_expr, expr_attr_names, expr_attr_values):
    return encode_expressions(None, None, None, None, projection_expr, expr_attr_names, expr_attr_values).Projection

def encode_expressions(condition_expr, key_condition_expr, filter_expr, update_expr, projection_expr,
        expr_attr_names, expr_attr_values):
    exprs = [
        ('Condition', condition_expr),
        ('KeyCondition', key_condition_expr),
        ('Filter', filter_expr),
        ('Update', update_expr),
        ('Projection', projection_expr)
    ]

    output = Expressions()

    generator = CborSExprGenerator(expr_attr_names, expr_attr_values)

    for _type, expr in exprs:
        if not expr:
            setattr(output, _type, None)
            continue

        typestr = _type + 'Expression'
        if _type in ('Condition', 'KeyCondition', 'Filter'):
            expr_array_len = 3
            tree = DynamoDbExpressionParser.parse_condition(expr, ExpressionErrorListener(expr, typestr))
        elif _type == 'Projection':
            expr_array_len = 2
            tree = DynamoDbExpressionParser.parse_projection(expr, ExpressionErrorListener(expr, typestr))
        elif _type == 'Update':
            expr_array_len = 3
            tree = DynamoDbExpressionParser.parse_update(expr, ExpressionErrorListener(expr, typestr))
        else:
            raise DaxClientError('Unknown expresion type ' + str(_type), DaxErrorCode.Validation)

        generator.reset(_type)
        antlr4.tree.Tree.ParseTreeWalker.DEFAULT.walk(generator, tree)

        generator.validate_intermediate_state()
        spec = generator.stack.pop()

        enc = CborEncoder()
        enc.append_array_header(expr_array_len)
        enc.append_int(CborSExprGenerator.ENCODING_FORMAT)
        enc.append_raw(spec)

        if _type != 'Projection':
            enc.append_array_header(len(generator.var_values))
            for var_val in generator.var_values:
                enc.append_raw(var_val)

        setattr(output, _type, enc.as_bytes())

    generator.validate_final_state()

    output.ExpressionAttributeNames = expr_attr_names
    output.ExpressionAttributeValues = expr_attr_values

    return output

class CborSExprGenerator(DynamoDbGrammarListener):
    ENCODING_FORMAT = 1
    ATTRIBUTE_VALUE_PREFIX = ':'
    ATTRIBUTE_NAME_PREFIX = '#'

    def __init__(self, expression_attribute_names, expression_attribute_values):
        self.stack = []
        self.reset(None)

        self.expr_attr_names = expression_attribute_names
        self.expr_attr_values = expression_attribute_values

        self.unused_expr_attr_names = set(self.expr_attr_names.keys()) if self.expr_attr_names else set()
        self.unused_expr_attr_values = set(self.expr_attr_values.keys()) if self.expr_attr_values else set()

    def reset(self, _type):
        self._type = _type
        self.nesting_level = 0
        self.var_name_by_id = {}
        self.var_values = []

    def validate_intermediate_state(self):
        if len(self.stack) != 1:
            raise DaxValidationError('Invalid {}Expression, Stack size = {}'.format(self._type, len(self.stack)))

        if self.nesting_level != 0:
            raise DaxValidationError('Invalid {}Expression, Nesting level = {}'.format(self._type, self.nesting_level))

    def validate_final_state(self):
        if self.unused_expr_attr_names:
            names = self._join_missing_names(self.unused_expr_attr_names)
            raise DaxValidationError('Value provided in ExpressionAttributeNames unused in expressions: keys: {' + names + '}')

        if self.unused_expr_attr_values:
            names = self._join_missing_names(self.unused_expr_attr_values)
            raise DaxValidationError('Value provided in ExpressionAttributeValues unused in expressions: keys: {' + names + '}')

    def _validate_not_equals(self, expr_type, actual, not_expected):
        for n in not_expected:
            if actual.lower() == n.lower():
                exp_type_str = expr_type or ''
                raise DaxValidationError(
                    "Invalid {}Expression: The function '{}' is not allowed in a {} expression".format(
                        exp_type_str, actual, exp_type_str.lower()))

    def _join_missing_names(self, names):
        return ', '.join(names)

    def enterComparator(self, ctx):
        self.nesting_level += 1

    def exitComparator(self, ctx):
        arg2 = self.stack.pop()
        func = self.stack.pop()
        arg1 = self.stack.pop()

        self.stack.append(self._encode_array([func, arg1, arg2]))
        self.nesting_level -= 1

    def exitComparator_symbol(self, ctx):
        text = ctx.getText()
        try:
            func = OPERATORS[text]
        except KeyError:
            raise DaxClientError('Invalid function ' + text, DaxErrorCode.Validation, False)

        self.stack.append(self._encode_function_code(func))

    def exitPath(self, ctx):
        n = ctx.getChildCount()
        components = self.stack[-n:]; del self.stack[-n:]

        self.stack.append(self._encode_function(Func.DocumentPath, components))

    def exitListAccess(self, ctx):
        value = ctx.getText()
        ordinal = int(value[1:-1]) # get rid of []
        self.stack.append(self._encode_list_access(ordinal))

    def exitId_(self, ctx):
        _id = ctx.getText()
        if _id[0] == CborSExprGenerator.ATTRIBUTE_NAME_PREFIX:
            try:
                sub = self.expr_attr_names[_id]
            except KeyError:
                raise DaxValidationError('Invalid {}Expression. Substitution value not provided for {}'.format(self._type, _id))

            self.unused_expr_attr_names.discard(_id)
            self.stack.append(self._encode_attribute_value({'S': sub}))
        else:
            self.stack.append(self._encode_document_path_element(_id))

    def exitLiteralSub(self, ctx):
        literal = ctx.getText()
        self.stack.append(self._encode_variable(literal[1:]))

    def exitAnd(self, ctx):
        arg2 = self.stack.pop()
        arg1 = self.stack.pop()
        self.stack.append(self._encode_function(Func.And, [arg1, arg2]))

    def exitOr(self, ctx):
        arg2 = self.stack.pop()
        arg1 = self.stack.pop()
        self.stack.append(self._encode_function(Func.Or, [arg1, arg2]))

    def exitNegation(self, ctx):
        arg = self.stack.pop()
        self.stack.append(self._encode_function(Func.Not, [arg]))

    def enterIn(self, ctx):
        self.nesting_level += 1

    def exitIn(self, ctx):
        numargs = (ctx.getChildCount() - 3) // 2    # arg + IN + ( + args*2-1 + )
        args = self.stack[-numargs:]; del self.stack[-numargs:]
        arg1 = self.stack.pop()

        # a in (b,c,d) =>  (In a (b c d))
        self.stack.append(self._encode_function(Func.In, [arg1, self._encode_array(args)]))
        self.nesting_level -= 1

    def enterBetween(self, ctx):
        self.nesting_level += 1

    def exitBetween(self, ctx):
        args = self.stack[-3:]; del self.stack[-3:]

        # a between b and c => (Between a b c)
        self.stack.append(self._encode_function(Func.Between, args))
        self.nesting_level -= 1

    def enterFunctionCall(self, ctx):
        funcname = ctx.ID().getText()
        if self._type == 'Update':
            self._validate_not_equals(self._type, funcname, ['attribute_exists', 'attribute_not_exists',
                'attribute_type', 'begins_with', 'contains', 'size'])
            if self.nesting_level > 0 and funcname.lower() != 'if_not_exists':
                raise DaxValidationError('Only if_not_exists() function can be nested (got ' + funcname.lower() + ')')

        elif self._type == 'Filter' or self._type == 'Condition':
            self._validate_not_equals(self._type, funcname, ['if_not_exists', 'list_append'])
            if self.nesting_level == 0 and funcname.lower() == 'size':
                raise DaxValidationError(
                        "Invalid {}Expression: The function '{}' is not allowed to be used this way in an expression".format(
                    self._type, funcname))
            elif self.nesting_level > 0 and funcname.lower() != 'size':
                raise DaxValidationError('Only size() function can be nested (got ' + funcname.lower() + ')')

        self.nesting_level += 1

    def exitFunctionCall(self, ctx):
        funcname = ctx.ID().getText().lower()

        try:
            func = FUNCS[funcname]
        except KeyError:
            raise DaxValidationError(
                    'Invalid {}Expression: Invalid function name: function: {}'.format(self._type, funcname))

        numargs = (ctx.getChildCount() - 2) // 2     # children = fname + ( + numOperands*2-1 +)
        args = self.stack[-numargs:]; del self.stack[-numargs:]

        # func(a,b,c,..) => (func a b c ..)
        self.stack.append(self._encode_function(func, args))
        self.nesting_level -= 1

    def exitProjection(self, ctx):
        numpaths = (ctx.getChildCount() + 1) // 2 # path, path, ... path
        paths = self.stack[-numpaths:]; del self.stack[-numpaths:]
        self.stack.append(self._encode_array(paths))

    def exitUpdate(self, ctx):
        updates = self.stack[:]; del self.stack[:]
        self.stack.append(self._encode_array(updates))

    def exitSet_action(self, ctx):
        operand = self.stack.pop()
        path = self.stack.pop()
        self.stack.append(self._encode_function(Func.SetAction, [path, operand]))

    def exitRemove_action(self, ctx):
        path = self.stack.pop()
        self.stack.append(self._encode_function(Func.RemoveAction, [path]))

    def exitAdd_action(self, ctx):
        value = self.stack.pop()
        path = self.stack.pop()
        self.stack.append(self._encode_function(Func.AddAction, [path, value]))

    def exitDelete_action(self, ctx):
        value = self.stack.pop()
        path = self.stack.pop()
        self.stack.append(self._encode_function(Func.DeleteAction, [path, value]))

    def enterPlusMinus(self, ctx):
        self.nesting_level += 1

    def exitPlusMinus(self, ctx):
        op2 = self.stack.pop()
        op1 = self.stack.pop()

        operator = ctx.getChild(1).getText()
        try:
            func = OPERATORS[operator]
        except KeyError:
            raise DaxClientError('Must be +/-', DaxErrorCode.Validation, False)

        self.stack.append(self._encode_function(func, [op1, op2]))
        self.nesting_level -= 1

    def _encode_document_path_element(self, s):
        return CborEncoder().append_string(s).as_bytes()

    def _encode_attribute_value(self, val):
        return AttributeValueEncoder.encode_attribute_value(val)

    def _encode_array(self, array):
        enc = CborEncoder()
        enc.append_array_header(len(array))
        for item in array:
            # Array items are already CBOR encoded
            enc.append_raw(item)
        return enc.as_bytes()

    def _encode_function_code(self, func):
        return CborEncoder().append_int(func).as_bytes()

    def _encode_function(self, func, args):
        enc = CborEncoder()
        enc.append_array_header(len(args) + 1)
        enc.append_int(func)
        for arg in args:
            # Args are already CBOR encoded
            enc.append_raw(arg)
        return enc.as_bytes()

    def _encode_list_access(self, ordinal):
        enc = CborEncoder()
        enc.append_tag(DaxCborTypes.TAG_DDB_DOCUMENT_PATH_ORDINAL)
        enc.append_int(ordinal)
        return enc.as_bytes()

    def _encode_variable(self, var_name):
        fullname = CborSExprGenerator.ATTRIBUTE_VALUE_PREFIX + var_name
        try:
            val = self.expr_attr_values[fullname]
        except KeyError:
            raise DaxClientError(
                'Invalid {}Expression: An expression attribute value used in expression is not defined: {}'.format(
                    self._type, fullname),
                DaxErrorCode.Validation, False)

        self.unused_expr_attr_values.discard(fullname)
        try:
            var_id = self.var_name_by_id[var_name]
        except KeyError:
            var_id = len(self.var_values)
            self.var_name_by_id[var_name] = var_id
            self.var_values.append(self._encode_attribute_value(val))

        enc = CborEncoder()
        enc.append_array_header(2)
        enc.append_int(Func.Variable)
        enc.append_int(var_id)
        return enc.as_bytes()

from antlr4.error.ErrorListener import ErrorListener
class ExpressionErrorListener(ErrorListener):
    def __init__(self, expr, expr_type):
        super(ExpressionErrorListener, self).__init__()
        self.expr = expr
        self.expr_type = expr_type

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise DaxClientError(
            'Invalid {}: Syntax error; token: "{}", near: line {} char {}'.format(self.expr_type, offendingSymbol.text, line, column),
            DaxErrorCode.Validation, False)

class Func:
    # NOTE = Ordinal is used as identifiers in CBor encoded format
    # Comparison operators #
    Equal = 0
    NotEqual = 1
    LessThan = 2
    GreaterEqual = 3
    GreaterThan = 4
    LessEqual = 5

    # Logical operators #
    And = 6
    Or = 7
    Not = 8

    # Range operators #
    Between = 9

    # Enumeration operators #
    In = 10

    # Functions #
    AttributeExists = 11
    AttributeNotExists = 12
    AttributeType = 13
    BeginsWith = 14
    Contains = 15
    Size = 16

    # Document path elements #
    Variable = 17       # takes 1 argument which is a placeholder for a value. function substitutes argument with corresponding value
    DocumentPath = 18   # maps a CBOR object to a document path

    # Update Actions #
    SetAction = 19
    AddAction = 20
    DeleteAction = 21
    RemoveAction = 22

    # Update operations #
    IfNotExists = 23
    ListAppend = 24
    Plus = 25
    Minus = 26

OPERATORS = {
    '+': Func.Plus,
    '-': Func.Minus,
    '=': Func.Equal,
    '<>': Func.NotEqual,
    '<': Func.LessThan,
    '<=': Func.LessEqual,
    '>': Func.GreaterThan,
    '>=': Func.GreaterEqual,
    'and': Func.And,
    'or': Func.Or,
    'not': Func.Not,
    'between': Func.Between,
    'in': Func.In,
    ':': Func.Variable,
    '.': Func.DocumentPath,
}

ACTIONS = {
    'SET': Func.SetAction,
    'ADD': Func.AddAction,
    'DELETE': Func.DeleteAction,
    'REMOVE': Func.RemoveAction,
}

FUNCS = {
    'attribute_exists': Func.AttributeExists,
    'attribute_not_exists': Func.AttributeNotExists,
    'attribute_type': Func.AttributeType,
    'begins_with': Func.BeginsWith,
    'contains': Func.Contains,
    'size': Func.Size,
    'if_not_exists': Func.IfNotExists,
    'list_append': Func.ListAppend
}



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import struct
import binascii
import decimal
import six

from . import CborTypes, compat

MAX_2 = 2 ** 16 - 1
MAX_4 = 2 ** 32 - 1
MAX_8 = 2 ** 64 - 1
MIN_8 = -MAX_8 - 1

# Taken from boto3 for compatibility
DYNAMODB_CONTEXT = decimal.Context(
    Emin=-128, Emax=126, prec=38,
    traps=[decimal.Clamped, decimal.Overflow, decimal.Inexact, decimal.Rounded, decimal.Underflow])

class CborEncoder(object):
    def __init__(self):
        # Preallocate common buffers
        self.buf_2 = bytearray(2)
        self.buf_4 = bytearray(4)
        self.buf_8 = bytearray(8)

        self._buffer = bytearray()

    def buffer(self):
        return self._buffer

    def as_bytes(self):
        return bytes(self._buffer)

    def reset(self):
        # TODO More efficient buffer handling, with pointers.
        self._buffer = bytearray()

    def _drain(self, n):
        ''' Remove the first n bytes from the buffer. '''
        del self._buffer[:n]

    def append_raw(self, data):
        self._buffer += data
        return self

    def append_int(self, n):
        if isinstance(n, six.integer_types):
            if n >= 0:
                if n <= MAX_8:
                    self._append_type(CborTypes.TYPE_POSINT, n)
                else:
                    self._append_bigint(CborTypes.TAG_POSBIGINT, n)
            else:
                if n >= MIN_8:
                    self._append_type(CborTypes.TYPE_NEGINT, -n - 1)
                else:
                    self._append_bigint(CborTypes.TAG_NEGBIGINT, -n - 1)
        else:
            raise TypeError('Expected {}, got {}'.format(six.integer_types, type(n).__name__))

        return self

    def append_float(self, f):
        self._append_type(CborTypes.TYPE_FLOAT_64)
        self._append_double(f)
        return self

    def append_decimal(self, d):
        if not isinstance(d, decimal.Decimal):
            raise TypeError('Decimals must be decimal.Decimal; got ' + type(d).__name__)

        self.append_tag(CborTypes.TAG_DECIMAL)
        self.append_array_header(2)

        exponent, mantissa = _crack_decimal(d)
        self.append_int(exponent)
        self.append_int(mantissa)

        return self

    def append_number(self, n):
        if isinstance(n, bytes):
            n = six.text_type(n, 'utf8')
        elif isinstance(n, bytearray):
            # Python 2 compat - can't decode bytearray with unicode()
            n = n.decode('utf8')

        if isinstance(n, six.text_type):
            if _is_decimal(n):
                n = DYNAMODB_CONTEXT.create_decimal(n)
            else:
                n = compat.bigint(n)

        if isinstance(n, six.integer_types):
            return self.append_int(n)
        elif isinstance(n, decimal.Decimal):
            return self.append_decimal(n)
        elif isinstance(n, float):
            return self.append_float(n)
        else:
            raise TypeError('Numbers must be strings, bytes, or a number type; got ' + type(n).__name__)

    def append_string(self, s):
        if not isinstance(s, six.string_types):
            raise TypeError("String must be of type {}, got {}".format(six.string_types, type(s).__name__))

        if not isinstance(s, six.text_type) and isinstance(s, bytes):
            # If we're passed a str object on Python 2, try converting it to
            # Unicode, while letting any UnicodeDecodeError exceptions bubble
            # up.
            _ = six.text_type(s)
            b = s
        else:
            b = compat.to_bytes(s, 'utf8')

        self._append_type(CborTypes.TYPE_UTF, len(b))
        self._buffer += b
        return self

    def append_binary(self, b):
        self._append_type(CborTypes.TYPE_BYTES, len(b))
        if isinstance(b, six.text_type):
            buf = compat.to_bytes(b, 'utf8')
        elif isinstance(b, (bytes, bytearray)):
            buf = b
        else:
            raise TypeError('Binary data must be {}, bytes, or bytearray; got {}'.format(
                six.text_type, type(b).__name__))

        self._buffer += buf
        return self

    def append_boolean(self, b):
        self._buffer.append(CborTypes.TYPE_TRUE if b else CborTypes.TYPE_FALSE)
        return self

    def append_null(self):
        self._buffer.append(CborTypes.TYPE_NULL)
        return self

    def append_break(self):
        self._buffer.append(CborTypes.TYPE_BREAK)
        return self

    def append_array_header(self, size):
        self._append_type(CborTypes.TYPE_ARRAY, size)
        return self

    def append_map_header(self, size):
        self._append_type(CborTypes.TYPE_MAP, size)
        return self

    def append_string_stream_header(self):
        self._buffer.append(CborTypes.TYPE_UTF_STREAM)
        return self

    def append_byte_stream_header(self):
        self._buffer.append(CborTypes.TYPE_BYTES_STREAM)
        return self

    def append_array_stream_header(self):
        self._buffer.append(CborTypes.TYPE_ARRAY_STREAM)
        return self

    def append_map_stream_header(self):
        self._buffer.append(CborTypes.TYPE_MAP_STREAM)
        return self

    def append_tag(self, tag):
        self._append_type(CborTypes.TYPE_TAG, tag)
        return self

    def _append_type(self, major_type, val=0):
        if not isinstance(val, six.integer_types):
            raise TypeError('Type value must be an integer, got ' + type(val).__name__)

        if val == 0:
            self._buffer.append(major_type)
        elif val < 24:
            self._buffer.append(major_type | val)
        elif val <= 0xff:
            self._buffer.append(major_type | CborTypes.SIZE_8)
            self._buffer.append(val)
        elif val <= MAX_2:
            self._buffer.append(major_type | CborTypes.SIZE_16)
            self._append_uint16be(val)
        elif val <= MAX_4:
            self._buffer.append(major_type | CborTypes.SIZE_32)
            self._append_uint32be(val)
        elif val <= MAX_8:
            self._buffer.append(major_type | CborTypes.SIZE_64)
            self._append_uint64be(val)
        else:
            raise ValueError('Type value must be <= 2^64, got ' + str(val))

    def _append_bigint(self, tag, val):
        self.append_tag(tag)

        # There may be faster ways than using a string conversion but bigint
        # encoding is unlikely to be a bottleneck
        hval = '%x' % val
        bval = binascii.unhexlify('0' + hval if len(hval) % 2 != 0 else hval)

        self.append_binary(bval)

    def _append_uint16be(self, val):
        struct.pack_into("!H", self.buf_2, 0, val)
        self._buffer += self.buf_2

    def _append_uint32be(self, val):
        struct.pack_into("!I", self.buf_4, 0, val)
        self._buffer += self.buf_4

    def _append_uint64be(self, val):
        struct.pack_into("!Q", self.buf_8, 0, val)
        self._buffer += self.buf_8

    def _append_float(self, val):
        struct.pack_into('!f', self.buf_4, 0, val)
        self._buffer += self.buf_4

    def _append_double(self, val):
        struct.pack_into('!d', self.buf_8, 0, val)
        self._buffer += self.buf_8

def _is_decimal(s):
    for c in s:
        if c == '.' or c == 'e' or c == 'E':
            return True

    return False

def _crack_decimal(d):
    '''
    Oddly enough, this is the fastest method in Python, because as_tuple() is
    so slow:

    In [5]: timeit('str(d)', globals=globals())
    Out[5]: 0.44962017599027604

    In [6]: timeit('d.as_tuple()', globals=globals())
    Out[6]: 1.1225044820166659

    Code originally from https://stackoverflow.com/a/24945278/129592 & added
    engineering notation handling & proper 0 handling.
    '''
    base_exp = 0
    s = str(d)
    eng_parts = s.split('E')

    if len(eng_parts) == 2:
        # We have engineering notation
        base_exp = int(eng_parts[1])
        s = eng_parts[0]

    parts = s.split('.')

    if len(parts) == 2:
        # Negative exponent
        exponent = -len(parts[1])
        mantissa = int(parts[0] + parts[1])
    else:
        # Positive exponent
        digits = parts[0].rstrip('0')
        digits = digits if digits != '' else '0' # Handle 0 case
        exponent = len(parts[0]) - len(digits)
        mantissa = int(digits)

    return base_exp+exponent, mantissa



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from .CborDecoder import CborDecoder
from .CborTypes import TYPE_ARRAY, is_major_type, major_type_name
from .DaxCborTypes import (TAG_DDB_STRING_SET, TAG_DDB_NUMBER_SET,
                           TAG_DDB_BINARY_SET, TAG_DDB_DOCUMENT_PATH_ORDINAL,
                           DdbSet, DocumentPathOrdinal)
from .DaxError import DaxClientError, DaxErrorCode

class DaxCborDecoder(CborDecoder):
    def __init__(self, more):
        super(DaxCborDecoder, self).__init__(more, DAX_TAG_HANDLERS)

def _decode_set(dec, tag):
    t = dec.peek()
    if not is_major_type(t, TYPE_ARRAY):
        raise DaxClientError('Type for Set is not array: ' + major_type_name(t), DaxErrorCode.Decoder)

    if tag == TAG_DDB_STRING_SET:
        set_type = 'SS'
        values = [_dec.decode_string() for _dec in dec.decode_array_iter()]
    elif tag == TAG_DDB_NUMBER_SET:
        set_type = 'NS'
        values = [str(_dec.decode_number()) for _dec in dec.decode_array_iter()]
    elif tag == TAG_DDB_BINARY_SET:
        set_type = 'BS'
        values = [_dec.decode_binary() for _dec in dec.decode_array_iter()]
    else:
        raise DaxClientError('Invalid tag to decode set: ' + hex(tag), DaxErrorCode.Decoder)

    return DdbSet(set_type, values)

def _decode_document_path_ordinal(dec, tag):
    if tag != TAG_DDB_DOCUMENT_PATH_ORDINAL:
        raise DaxClientError('Invalid tag to decode document path ordinal: ' + hex(tag), DaxErrorCode.Decoder)

    ordinal = dec.decode_int()

    return DocumentPathOrdinal(ordinal)

DAX_TAG_HANDLERS = {
    TAG_DDB_STRING_SET: _decode_set,
    TAG_DDB_NUMBER_SET: _decode_set,
    TAG_DDB_BINARY_SET: _decode_set,
    TAG_DDB_DOCUMENT_PATH_ORDINAL: _decode_document_path_ordinal
}



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import random
import threading
import time
import socket
import itertools

from concurrent.futures import ThreadPoolExecutor, wait as future_wait
from collections import defaultdict, namedtuple

from . import ClusterUtil
from .DaxError import DaxClientError, DaxServiceError, DaxErrorCode
from .Assemblers import ENDPOINT_FIELDS

import logging
logger = logging.getLogger(__name__)

_ServiceEndpoint = namedtuple('_ServiceEndpoint', ENDPOINT_FIELDS)
AddrPort = namedtuple('AddrPort', ('address', 'port'))

class ServiceEndpoint(_ServiceEndpoint):
    @classmethod
    def from_endpoint(cls, endpoint):
        # Set the hostname to the address if none provided
        if 'hostname' in endpoint:
            if 'address' not in endpoint:
                # No address, so go ask DNS
                addresses = _resolve_dns(endpoint['hostname'], endpoint['port'])
                if addresses:
                    # If multiple addresses, pick one
                    endpoint['address'] = random.choice(addresses)[0]
                else:
                    # Could not resolve hostname, bad endpoint, ignore it
                    return None
        else:
            if endpoint.setdefault('hostname', endpoint.get('address')) is None:
                # No hostname or address, go to the next endpoint
                return None

        endpoint.setdefault('leader_session_id', 0)
        return cls(**endpoint)

    @classmethod
    def from_endpoints(cls, endpoints):
        ''' Create a frozenset of ServiceEndpoint objects from a list of endpoints.

        Endpoints without valid addresses are dropped.
        '''
        service_endpoints = [cls.from_endpoint(endpoint) for endpoint in endpoints]
        return frozenset(service_endpoint for service_endpoint in service_endpoints if service_endpoint is not None)

    @property
    def addrport(self):
        try:
            return self._addrport
        except AttributeError:
            self._addrport = AddrPort(self.address, self.port) # pylint: disable=attribute-defined-outside-init
            return self._addrport

class Role(object):
    LEADER = 1
    REPLICA = 2

    @staticmethod
    def name(n):
        if n == Role.LEADER:
            return "LEADER"
        elif n == Role.REPLICA:
            return "REPLICA"
        else:
            raise ValueError("Unknown role %d" % n)

class Router(object):
    ''' Abstract class that determines how the cluster routes requests. '''
    def close(self):
        raise NotImplementedError()

    def start(self):
        raise NotImplementedError()

    def update(self, service_endpoints):
        raise NotImplementedError()

    def wait_for_routes(self, min_healthy=1, leader_min=1, timeout=None):
        raise NotImplementedError()

    def next_leader(self, prev_client):
        raise NotImplementedError()

    def next_any(self, prev_client):
        raise NotImplementedError()

class EndpointRouter(Router):
    ''' Determine which nodes to route requests to.

    This manages the EndpointBackend objects that represent nodes in the cluster, and allows retrieving the appropriate
    node type (leader or replica).
    '''
    def __init__(self,
                 client_factory,
                 discovery_endpoints,
                 update_cluster_interval,
                 health_check_interval,
                 selector=None,
                 executor=None):
        self._client_factory = client_factory
        self.selector = selector or random_backend

        self._discovery_endpoints = discovery_endpoints

        # endpoints returned from last successful endpoints() call to the cluster
        self._service_endpoints = frozenset()

        # backends is a dict[role, dict[(hostname,port,node), backend] of the
        # active backends
        self.backends = defaultdict(dict)

        # dict of AddrPort -> EndpointBackend of inactive backeends. When they
        # come up, they are moved to self.backends.
        self._pending_backends = {}

        self._health_check_interval = health_check_interval
        self._update_cluster_interval = update_cluster_interval

        self._lock = threading.RLock()
        self._route_change = threading.Condition(self._lock)

        # Executor for health checks
        self._background = executor or ThreadPoolExecutor(max_workers=2)

        # Thread that refreshes the roster
        self._refresh_task = None

    @property
    def all_backends(self):
        return list(self._all_backends())

    def _all_backends(self):
        return itertools.chain.from_iterable(role_backends.values() for role, role_backends in self.backends.items())

    @property
    def leader_backends(self):
        return list(self.backends[Role.LEADER].values())

    def close(self):
        with self._lock:
            if self._refresh_task:
                self._refresh_task.cancel()
                self._refresh_task = None

            backends = self.all_backends + list(self._pending_backends.values())

        closers = [self._background.submit(backend.close) for backend in backends]
        done, not_done = future_wait(closers, timeout=5.0) # pylint: disable=unused-variable
        # TODO What to do with not_done?

        with self._lock:
            self.backends.clear()
            self.backends = None

        self._background.shutdown()

    def update(self, service_endpoints):
        ''' Update the set of connected endpoints.

        If an endpoint has changed roles, move it. If it is new, create a new backend.
        If a backend is not in the new endpoint list, close & remove it.
        '''
        if self._service_endpoints == service_endpoints:
            # If the backends haven't changed, skip the update process
            return

        pending = []
        with self._lock:
            for ep in service_endpoints:
                self._update_endpoint(ep)

            pending.extend(self._purge_endpoints(service_endpoints))
            self._service_endpoints = service_endpoints

        done, not_done = future_wait(pending, timeout=5.0) # pylint: disable=unused-variable
        # TODO What to do with not_done?

        logger.debug('Current backends: %s', dict(self.backends))

    def _update_endpoint(self, service_endpoint):
        ''' Update any endpoints that have changed or are new.

        Not thread-safe. Must be called from inside update().
        '''
        new_addrport = service_endpoint.addrport
        new_role = service_endpoint.role

        backend = self.backends[new_role].get(new_addrport)
        if backend is not None:
            # Backend is up-to-date, no changes necessary
            logger.debug('Updating active backend %s', backend)
            backend.update(service_endpoint)
        else:
            # Role may have changed, so search other roles to see if we're already connected
            for role, role_backends in self.backends.items():
                if role == new_role:
                    # Already checked...
                    continue

                backend = role_backends.pop(new_addrport, None)
                if backend is not None:
                    # Backend has only changed roles, so just move it
                    backend.update(service_endpoint)
                    self.backends[new_role][new_addrport] = backend
                    logger.debug('%s changing role from %s to %s', backend, Role.name(role), Role.name(new_role))
                    break
            else: # no break
                # Update it if pending
                if new_addrport in self._pending_backends:
                    logger.debug('Updating pending backend %s', self._pending_backends[new_addrport])
                    self._pending_backends[new_addrport].update(service_endpoint)
                else:
                    # Backend does not exist, so create it
                    # It will add itself when it is ready
                    self._create_backend(service_endpoint)

    def _purge_endpoints(self, service_endpoints):
        ''' Remove backends that no longer have corresponding endpoints.

        Not thread-safe. Must be called from inside update().
        '''
        endpoints = {ep.addrport for ep in service_endpoints}

        pending = []
        for role_backends in self.backends.values():
            to_remove = []
            for addrport, backend in role_backends.items():
                if addrport not in endpoints:
                    # This backend is no longer in the service endpoints, so close it
                    # It will remove itself when closed
                    logger.debug('Removing unused backend %s', backend)
                    pending.append(self._background.submit(backend.close)) # Method ref, not call
                    to_remove.append(addrport) # Can't delete while looping

            # Remove all keys
            for addrport in to_remove:
                del role_backends[addrport]

        to_remove = []
        for addrport, backend in self._pending_backends.items():
            if addrport not in endpoints:
                logger.debug('Removing unused pending backend %s', addrport)
                pending.append(self._background.submit(backend.close))
                to_remove.append(addrport)

        for addrport in to_remove:
            del self._pending_backends[addrport]

        return pending

    def _create_backend(self, service_endpoint):
        ''' Create a new Backend instance and add it to the pending list.

        If a backend is already pending, do nothing.

        Not thread-safe. Must be called from inside update().
        '''

        logger.debug('Creating backend for %s', service_endpoint)

        pending_backend = self._pending_backends.get(service_endpoint.addrport)
        if pending_backend is None:
            logger.debug('Creating backend for %s', service_endpoint)
            pending_backend = EndpointBackend(
                service_endpoint,
                self._client_factory,
                self._health_check_interval,
                self._backend_up,
                self._backend_down)
            self._pending_backends[service_endpoint.addrport] = pending_backend

            # Bring the backend up in the background
            self._background.submit(pending_backend.start)
        else:
            # There is already a pending backend for this endpoint,
            # so let it continue instead
            pass

    def _backend_up(self, backend):
        with self._lock:
            self.backends[backend.role][backend.addrport] = backend
            self._pending_backends.pop(backend.addrport, None)
            self._route_change.notify_all()
        logger.debug('Backend up: %s', backend)

    def _backend_down(self, backend):
        role = backend.role
        with self._lock:
            # Move it to pending if it's being brought down because of a health
            # check or other failure. Otherwise it's been removed from the
            # roster.
            if next((e for e in self._service_endpoints if e.addrport == backend.addrport), False):
                self._pending_backends[backend.addrport] = backend
            self.backends[role].pop(backend.addrport, None)
            self._route_change.notify_all()
        logger.debug('Backend down: %s', backend)

    def wait_for_routes(self, min_healthy=1, leader_min=1, timeout=None):
        start = _clock()
        while True:
            with self._lock:
                if self._has_active_any(min_healthy) and self._has_active_leaders(leader_min):
                    return
                else:
                    now = _clock()
                    if timeout is not None and now - start > timeout:
                        raise DaxClientError(
                            "Not enough routes after {}s: expected {}/{}, found {}/{} (healthy/leaders)".format(
                                timeout, min_healthy, leader_min, self._active_any(), self._active_leaders()),
                            DaxErrorCode.NoRoute)

                # Wait for a signal that the endpoints have changed
                self._route_change.wait(timeout)

    def _active_any(self):
        return sum(len(role_backend) for role, role_backend in self.backends.items())

    def _has_active_any(self, min_healthy):
        if min_healthy < 1:
            return True
        else:
            return self._active_any() >= min_healthy

    def _active_leaders(self):
        return len(self.backends[Role.LEADER])

    def _has_active_leaders(self, leader_min):
        if leader_min < 1:
            return True
        else:
            return self._active_leaders() >= leader_min

    def next_leader(self, prev_client):
        ''' Returns the next leader entry that is not the given prev value if one such entry is available.

        If there is only one entry and that is equals to prev, prev is returned.
        Returns None if nothing is available.
        '''

        with self._lock:
            if self.backends is not None:
                backend = self.selector(prev_client, self.leader_backends)
                return backend.client if backend is not None else None

    def next_any(self, prev_client):
        ''' Returns any entry that is not the given prev value, if any such entry is available.

        If there is only one entry and that is equals to prev, prev is returned.
        Returns None if nothing is available.
        '''
        with self._lock:
            if self.backends is not None:
                backend = self.selector(prev_client, self.all_backends)
                return backend.client if backend is not None else None

    def start(self):
        ''' Set up the initial roster of nodes using the provided discovery endpoints. '''

        self._bootstrap()
        if not self._service_endpoints:
            raise DaxClientError('Failed to configure cluster endpoints from {}'.format(self._discovery_endpoints), DaxErrorCode.NoRoute)

        with self._lock:
            if self._refresh_task:
                self._refresh_task.cancel()

            self._refresh_task = ClusterUtil.periodic_task(self._bootstrap_async, self._update_cluster_interval, jitter=0.5)

    def _bootstrap_async(self):
        '''Non-throwing bootstrap, for calling from a background thread.'''

        try:
            self._bootstrap()
        except Exception as e: # pylint: disable=broad-except
            logger.warning('Failed to boostrap endpoints', exc_info=e)

    def _bootstrap(self):
        service_endpoints = self._bootstrap_endpoints()
        logger.debug('Bootstrapped endpoints: %s', service_endpoints)
        if service_endpoints:
            self.update(service_endpoints)

    def _bootstrap_endpoints(self):
        logger.debug("Bootstrapping from discovery endpoints %s", self._discovery_endpoints)

        seeds = []
        for host, port in self._discovery_endpoints:
            # Find all IP addresses for this hostname from DNS
            seeds_from_endpoint = _resolve_dns(host, port)
            logger.debug('Resolved addresses %s for %s', seeds_from_endpoint, host)
            seeds += seeds_from_endpoint

        random.shuffle(seeds)
        service_endpoints = self._get_endpoints_from_seeds(seeds)
        if service_endpoints:
            return service_endpoints

    def _get_endpoints_from_seeds(self, seeds):
        for ip, port in seeds:
            # Stop after the first successful call
            try:
                client = self._client_factory(ip, port)
            except Exception as e: # pylint: disable=broad-except
                logger.info('Could not connect to %s:%s', ip, port, exc_info=True)
                continue
            else:
                try:
                    endpoints = client.endpoints()
                except DaxServiceError as e:
                    if e.auth_error:
                        # Startup will fail with authentication errors if region
                        # is not set correctly on client config. Region can also be
                        # set after instantiating cluster client. Ignore the exception.
                        logger.warning('Auth exception while starting up cluster client: %s', e)

                        # Re-raise because this will be unrecoverable
                        raise
                except Exception as e: # pylint: disable=broad-except
                    logger.info('Failed to retrieve endpoints from %s:%s', ip, port, exc_info=True)
                    continue
                else:
                    if endpoints:
                        return ServiceEndpoint.from_endpoints(endpoints)
                finally:
                    client.close()


def random_backend(prev_client, backends):
    ''' Select a random available client. '''
    if not backends:
        return None
    elif len(backends) == 1:
        return backends[0]
    else:
        # Select a random backend
        i = random.randrange(len(backends))
        be = backends[i]
        if be.client is prev_client:
            # Just pick the next one
            be = backends[(i + 1) % len(backends)]
        return be

class EndpointBackend(object):

    @staticmethod
    def _noop(*args, **kwargs):
        pass

    def __init__(self, service_endpoint, client_factory, health_check_interval, on_up, on_down):
        self._service_endpoint = service_endpoint
        self._client_factory = client_factory
        self._health_check_interval = health_check_interval
        self._connection_timeout = self._health_check_interval / 2

        self._on_up = on_up
        self._on_down = on_down

        self._error_count = 0

        self.client = None
        self.session = None
        self.active = False
        self._healthy = True  # Healthy until proven otherwise
        self._closed = False

        self._lock = threading.RLock()
        self._timer = None

    def start(self):
        self._health_check()
        logger.debug('Creating health checker for %s', self)
        self._timer = ClusterUtil.periodic_task(
            self._health_check,
            self._health_check_interval,
            self._health_check_interval * 0.1)

    def update(self, new_service_endpoint):
        if self._service_endpoint != new_service_endpoint:
            if self._service_endpoint.addrport != new_service_endpoint.addrport:
                raise ValueError('Cannot update backend to new address.')

            self._service_endpoint = new_service_endpoint
            return True

        return False

    def close(self):
        with self._lock:
            if self._closed:
                return

            logger.debug('Closing %s', self)

            if self._timer:
                self._timer.cancel()
                self._timer = None

            if self.active:
                self.down()

            self._closed = True

    @property
    def addrport(self):
        return self._service_endpoint.addrport

    @property
    def role(self):
        return self._service_endpoint.role

    @property
    def leader(self):
        return self._service_endpoint.role == Role.LEADER

    @property
    def suspect(self):
        return self._error_count > 0

    def up(self):
        upped = False
        with self._lock:
            if not self.active:
                self._error_count = 0
                self.active = True
                upped = True

        if upped:
            self._on_up(self)

    def down(self):
        if self._closed:
            return

        self._on_down(self)

        with self._lock:
            self.active = False
            if self.client:
                self.client.close()
                self.client = None

    def _health_check(self):
        with self._lock:
            if self._closed:
                return

            if not self.client:
                self.client = self._client_factory(*self.addrport)

            # Determine health by calling endpoints
            try:
                logger.debug('Running health check for %s', self)
                _ = self.client.endpoints()
            except Exception as e: # pylint: disable=broad-except
                logger.warning("Health check failed for %s: %s", self.addrport.address, e)
                if self.active:
                    self.down()

                return
            else:
                if not self.active:
                    self.up()

        return

    def __repr__(self):
        return 'EndpointBackend({})'.format(self._service_endpoint)

def _clock():
    return time.time()

def _resolve_dns(host, port):
    try:
        # Deliberately restrict it to IPv4 addresses
        return [sockaddr \
                for family, socktype, proto, canonname, sockaddr \
                in socket.getaddrinfo(host, port, socket.AF_INET, 0, socket.IPPROTO_TCP)]
    except socket.gaierror:
        # if there is an error, return no addresses
        return []
